/*
 * CDDL HEADER START
 *
 * The contents of this file are subject to the terms of the
 * Common Development and Distribution License (the "License").
 * You may not use this file except in compliance with the License.
 *
 * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
 * or http://www.opensolaris.org/os/licensing.
 * See the License for the specific language governing permissions
 * and limitations under the License.
 *
 * When distributing Covered Code, include this CDDL HEADER in each
 * file and include the License file at usr/src/OPENSOLARIS.LICENSE.
 * If applicable, add the following below this CDDL HEADER, with the
 * fields enclosed by brackets "[]" replaced with your own identifying
 * information: Portions Copyright [yyyy] [name of copyright owner]
 *
 * CDDL HEADER END
 */
/*
 * Copyright 2009 Sun Microsystems, Inc.  All rights reserved.
 * Use is subject to license terms.
 */

#include <libintl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <stddef.h>
#include <string.h>
#include <stropts.h>
#include <libdevinfo.h>
#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_types.h>
#include <libdlpi.h>
#include <libdllink.h>
#include <libdlwlan.h>
#include <libdladm_impl.h>
#include <libdlwlan_impl.h>

static int wpa_instance_create(char *ifname);
static int wpa_instance_delete(char *ifname);
static int wpa_network_config(char *ifname, dladm_wlan_attr_t *attrp,
	dladm_wlan_key_t *keys, boolean_t *create_ibss, char *identity);

static dladm_status_t 	do_get_bsstype(dladm_handle_t, datalink_id_t, void *,
			    int);
static dladm_status_t 	do_get_signal(dladm_handle_t, datalink_id_t, void *,
			    int);
static dladm_status_t 	do_get_encryption(dladm_handle_t, datalink_id_t, void *,
			    int);
static dladm_status_t 	do_get_authmode(dladm_handle_t, datalink_id_t, void *,
			    int);
static dladm_status_t 	do_get_linkstatus(dladm_handle_t, datalink_id_t, void *,
			    int);
static dladm_status_t 	do_get_rate(dladm_handle_t, datalink_id_t, void *, int);
static dladm_status_t	do_get_mode(dladm_handle_t, datalink_id_t, void *, int);
static dladm_status_t	do_get_wpamode(dladm_handle_t, datalink_id_t, void *,
			    int);
static dladm_status_t	do_set_key(dladm_handle_t, datalink_id_t,
			    dladm_wlan_key_t *, uint_t);
static dladm_status_t	do_scan(dladm_handle_t, datalink_id_t, void *, int);
static dladm_status_t	do_disconnect(dladm_handle_t, datalink_id_t, void *,
			    int);
static boolean_t	find_val_by_name(const char *, val_desc_t *,
			    uint_t, uint_t *);
static boolean_t	find_name_by_val(uint_t, val_desc_t *, uint_t, char **);
static void		generate_essid(dladm_wlan_essid_t *);

static dladm_status_t	dladm_wlan_wlresult2status(wldp_t *);
static dladm_status_t	dladm_wlan_validate(dladm_handle_t, datalink_id_t);

static val_desc_t	linkstatus_vals[] = {
	{ "disconnected", DLADM_WLAN_LINK_DISCONNECTED	},
	{ "connected",    DLADM_WLAN_LINK_CONNECTED	}
};

static val_desc_t 	secmode_vals[] = {
	{ "none",	DLADM_WLAN_SECMODE_NONE	},
	{ "wep",	DLADM_WLAN_SECMODE_WEP	},
	{ "wpa-psk",	DLADM_WLAN_SECMODE_PSK	},
	{ "wpa-eap",	DLADM_WLAN_SECMODE_EAP	}
};

static val_desc_t 	strength_vals[] = {
	{ "very weak",	DLADM_WLAN_STRENGTH_VERY_WEAK	},
	{ "weak",	DLADM_WLAN_STRENGTH_WEAK	},
	{ "good",	DLADM_WLAN_STRENGTH_GOOD	},
	{ "very good",	DLADM_WLAN_STRENGTH_VERY_GOOD	},
	{ "excellent",	DLADM_WLAN_STRENGTH_EXCELLENT	}
};

static val_desc_t	mode_vals[] = {
	{ "a",		DLADM_WLAN_MODE_80211A		},
	{ "b",		DLADM_WLAN_MODE_80211B		},
	{ "g",		DLADM_WLAN_MODE_80211G		},
	{ "n",		DLADM_WLAN_MODE_80211GN		},
	{ "n",		DLADM_WLAN_MODE_80211AN		}
};

/* capital chars*/
static val_desc_t	auth_vals[] = {
	{ "OPEN",	DLADM_WLAN_AUTH_OPEN		},
	{ "SHARED",	DLADM_WLAN_AUTH_SHARED		}
};

static val_desc_t	bsstype_vals[] = {
	{ "bss",	DLADM_WLAN_BSSTYPE_BSS		},
	{ "ibss",	DLADM_WLAN_BSSTYPE_IBSS		}
};

static dladm_status_t
dladm_wlan_wlresult2status(wldp_t *gbuf)
{
	switch (gbuf->wldp_result) {
	case WL_SUCCESS:
		return (DLADM_STATUS_OK);

	case WL_NOTSUPPORTED:
	case WL_LACK_FEATURE:
		return (DLADM_STATUS_NOTSUP);

	case WL_READONLY:
		return (DLADM_STATUS_PROPRDONLY);

	default:
		break;
	}

	return (DLADM_STATUS_FAILED);
}

static dladm_wlan_mode_t
do_convert_mode(wl_phy_conf_t *phyp)
{
	wl_erp_t *wlep = &phyp->wl_phy_erp_conf;
	wl_ofdm_t *wlop = &phyp->wl_phy_ofdm_conf;

	switch (phyp->wl_phy_fhss_conf.wl_fhss_subtype) {
	case WL_ERP:
		return (wlep->wl_erp_ht_enabled ?
		    DLADM_WLAN_MODE_80211GN : DLADM_WLAN_MODE_80211G);
	case WL_OFDM:
		return (wlop->wl_ofdm_ht_enabled ?
		    DLADM_WLAN_MODE_80211AN : DLADM_WLAN_MODE_80211A);
	case WL_DSSS:
	case WL_FHSS:
		return (DLADM_WLAN_MODE_80211B);
	default:
		break;
	}

	return (DLADM_WLAN_MODE_NONE);
}

boolean_t
i_dladm_wlan_convert_chan(wl_phy_conf_t *phyp, uint32_t *channelp)
{
	wl_fhss_t *wlfp = &phyp->wl_phy_fhss_conf;
	wl_ofdm_t *wlop = &phyp->wl_phy_ofdm_conf;

	switch (wlfp->wl_fhss_subtype) {
	case WL_FHSS:
	case WL_DSSS:
	case WL_IRBASE:
	case WL_HRDS:
	case WL_ERP:
		*channelp = wlfp->wl_fhss_channel;
		break;
	case WL_OFDM:
		*channelp = DLADM_WLAN_OFDM2CHAN(wlop->wl_ofdm_frequency);
		break;
	default:
		return (B_FALSE);
	}
	return (B_TRUE);
}

static void
fill_wlan_attr(wl_ess_conf_t *wlp, dladm_wlan_attr_t *attrp)
{
	int i;

	(void) memset(attrp, 0, sizeof (*attrp));

	attrp->wa_essid.we_length = wlp->wl_ess_conf_essid.wl_essid_length;
	(void) memcpy(attrp->wa_essid.we_bytes, wlp->wl_ess_conf_essid.wl_essid_essid,
	    attrp->wa_essid.we_length);
	attrp->wa_valid |= DLADM_WLAN_ATTR_ESSID;

	/*
	 * Note: SSID is an array of octets, i.e.,
	 * it is not nul terminated and can, at least in theory,
	 * contain control characters (including nul) and as such,
	 * should be processed as binary data, not a printable string.
	 *
	 *the following call has been replaced with the one above:
	 *(void) snprintf(attrp->wa_essid.we_bytes, DLADM_WLAN_MAX_ESSID_LEN,
	 *   "%s", wlp->wl_ess_conf_essid.wl_essid_essid);
	*/

	(void) memcpy(attrp->wa_bssid.wb_bytes, wlp->wl_ess_conf_bssid,
	    DLADM_WLAN_BSSID_LEN);
	attrp->wa_valid |= DLADM_WLAN_ATTR_BSSID;

	/*
	 * Open/Shared is simply not present in the AP's beacon.  It is a
	 * deficiency of WEP, which WPA fixed by only allowing open
	 * authentication for WPA connections. There is simply no way to tell
	 * whether an AP is using shared auth or open system auth because a WEP
	 * AP never broadcasts that information.
	 * [net80211 sets it to 1 (OPEN) by default for scanned nodes]
	 *
	 * If wpa_ie is not present and IEEE80211_CAP_PRIVACY is on => WEP
	 * If wpa_ie is not present and IEEE80211_CAP_PRIVACY is off => NONE
	 */

	if (wlp->wl_ess_conf_wpa_ie_len == 0) {
		if (wlp->wl_ess_conf_caps & IEEE80211_CAP_PRIVACY) {
			attrp->wa_auth = DLADM_WLAN_AUTH_SHARED;
			attrp->wa_secmode = DLADM_WLAN_SECMODE_WEP;
		} else {
			attrp->wa_auth = DLADM_WLAN_AUTH_NONE;
			attrp->wa_secmode = DLADM_WLAN_SECMODE_NONE;
		}
	} else {
		attrp->wa_auth = DLADM_WLAN_AUTH_OPEN;
		wpa_parse_wpa_ie(wlp->wl_ess_conf_wpa_ie,
		    wlp->wl_ess_conf_wpa_ie_len, &attrp->wa_ie);
		/*
		 * check wpa_ie.c defines
		 */
		if (attrp->wa_ie.key_mgmt == 0)
			attrp->wa_secmode = DLADM_WLAN_SECMODE_EAP;
		if (attrp->wa_ie.key_mgmt == 2)
			attrp->wa_secmode = DLADM_WLAN_SECMODE_PSK;
	}
	attrp->wa_valid |= DLADM_WLAN_ATTR_AUTH;
	attrp->wa_valid |= DLADM_WLAN_ATTR_SECMODE;

	/* caps values are not mutually-exclusive here */
	if (wlp->wl_ess_conf_caps & IEEE80211_CAP_IBSS)
	    	attrp->wa_bsstype = DLADM_WLAN_BSSTYPE_IBSS;
	if (wlp->wl_ess_conf_caps & IEEE80211_CAP_ESS)
	    	attrp->wa_bsstype = DLADM_WLAN_BSSTYPE_BSS;
	attrp->wa_valid |= DLADM_WLAN_ATTR_BSSTYPE;

	attrp->wa_strength = DLADM_WLAN_SIGNAL2STRENGTH(wlp->wl_ess_conf_sl);
	attrp->wa_valid |= DLADM_WLAN_ATTR_STRENGTH;

	attrp->wa_mode = do_convert_mode((wl_phy_conf_t *)&wlp->wl_phy_conf);
	attrp->wa_valid |= DLADM_WLAN_ATTR_MODE;

	for (i = 0; i < MAX_SCAN_SUPPORT_RATES; i++) {
		wlp->wl_supported_rates[i] &= IEEE80211_RATE;
		if (wlp->wl_supported_rates[i] > attrp->wa_speed)
			attrp->wa_speed = wlp->wl_supported_rates[i];
	}
	if (attrp->wa_speed > 0)
		attrp->wa_valid |= DLADM_WLAN_ATTR_SPEED;

	if (i_dladm_wlan_convert_chan((wl_phy_conf_t *)&wlp->wl_phy_conf,
	    &attrp->wa_channel))
		attrp->wa_valid |= DLADM_WLAN_ATTR_CHANNEL;
}

#define	WLDP_BUFSIZE	(MAX_BUF_LEN - WIFI_BUF_OFFSET)

dladm_status_t
dladm_wlan_scan(dladm_handle_t handle, datalink_id_t linkid, void *arg,
    boolean_t (*func)(void *, dladm_wlan_attr_t *))
{
	int			i;
	uint32_t		count;
	wl_ess_conf_t		*wlp;
	wl_ess_list_t		*wls = NULL;
	char 			buf[WLDP_BUFSIZE];
	wl_linkstatus_t		wl_status;
	dladm_wlan_attr_t	wlattr;
	dladm_status_t		status;

	if ((status = dladm_wlan_validate(handle, linkid)) != DLADM_STATUS_OK)
		goto done;

	status = do_get_linkstatus(handle, linkid, &wl_status,
	    sizeof (wl_status));
	if (status != DLADM_STATUS_OK)
		goto done;

	if ((status = do_scan(handle, linkid, buf, sizeof (buf))) !=
	    DLADM_STATUS_OK)
		goto done;

	if (func == NULL) {
		status = DLADM_STATUS_OK;
		goto done;
	}

	wls = malloc(WLDP_BUFSIZE);
	if (wls == NULL) {
		status = DLADM_STATUS_NOMEM;
		goto done;
	}

	if ((status = dladm_wlan_get_esslist(handle, linkid, wls, WLDP_BUFSIZE))
	    != DLADM_STATUS_OK)
		goto done;

	wlp = wls->wl_ess_list_ess;
	count = wls->wl_ess_list_num;

	for (i = 0; i < count; i++, wlp++) {
		fill_wlan_attr(wlp, &wlattr);
		if (!func(arg, &wlattr))
			break;
	}

	if (wl_status != WL_CONNECTED) {
		status = do_get_linkstatus(handle, linkid, &wl_status,
		    sizeof (&wl_status));
		if (status != DLADM_STATUS_OK)
			goto done;
		if (wl_status == WL_CONNECTED)
			(void) do_disconnect(handle, linkid, buf, sizeof (buf));
	}

	status = DLADM_STATUS_OK;
done:
	free(wls);
	return (status);
}

/*
 * Structures used in building the list of eligible WLANs to connect to.
 * Specifically, `connect_state' has the WLAN attributes that must be matched
 * (in `cs_attr') and a growing list of WLANs that matched those attributes
 * chained through `cs_list'.  Each element in the list is of type `attr_node'
 * and has the matching WLAN's attributes and a pointer to the next element.
 * For convenience, `cs_count' tracks the number of elements in the list.
 */
typedef struct attr_node {
	dladm_wlan_attr_t	an_attr;
	struct attr_node	*an_next;
} attr_node_t;

typedef struct connect_state {
	dladm_wlan_attr_t	*cs_attr;
	uint_t			cs_count;
	attr_node_t		*cs_list;
} connect_state_t;

/*
 * Compare two sets of WLAN attributes.  For now, we only consider strength
 * and speed (in that order), which matches the documented default policy for
 * dladm_wlan_connect().
 */
static int
attr_compare(const void *p1, const void *p2)
{
	dladm_wlan_attr_t *attrp1, *attrp2;

	attrp1 = (*(dladm_wlan_attr_t **)p1);
	attrp2 = (*(dladm_wlan_attr_t **)p2);

	if (attrp1->wa_strength < attrp2->wa_strength)
		return (1);

	if (attrp1->wa_strength > attrp2->wa_strength)
		return (-1);

	return (attrp2->wa_speed - attrp1->wa_speed);
}

/*
 * Callback function used by dladm_wlan_connect() to filter out unwanted
 * WLANs when scanning for available WLANs.  Always returns B_TRUE to
 * continue the scan.
 */
static boolean_t
connect_cb(void *arg, dladm_wlan_attr_t *attrp)
{
	attr_node_t		*nodep;
	dladm_wlan_attr_t	*fattrp;
	connect_state_t		*statep = (connect_state_t *)arg;

	fattrp = statep->cs_attr;
	if (fattrp == NULL)
		goto append;

	if ((fattrp->wa_valid & attrp->wa_valid) != fattrp->wa_valid)
		return (B_TRUE);

	if ((fattrp->wa_valid & DLADM_WLAN_ATTR_ESSID) != 0 &&
	    memcmp(fattrp->wa_essid.we_bytes, attrp->wa_essid.we_bytes,
	    attrp->wa_essid.we_length) != 0)
		return (B_TRUE);

	if ((fattrp->wa_valid & DLADM_WLAN_ATTR_SECMODE) != 0 &&
	    fattrp->wa_secmode != attrp->wa_secmode)
		return (B_TRUE);

	if ((fattrp->wa_valid & DLADM_WLAN_ATTR_MODE) != 0 &&
	    fattrp->wa_mode != attrp->wa_mode)
		return (B_TRUE);

	if ((fattrp->wa_valid & DLADM_WLAN_ATTR_STRENGTH) != 0 &&
	    fattrp->wa_strength != attrp->wa_strength)
		return (B_TRUE);

	if ((fattrp->wa_valid & DLADM_WLAN_ATTR_SPEED) != 0 &&
	    fattrp->wa_speed != attrp->wa_speed)
		return (B_TRUE);

	if ((fattrp->wa_valid & DLADM_WLAN_ATTR_AUTH) != 0) {
		attrp->wa_auth = fattrp->wa_auth;
		attrp->wa_valid |= DLADM_WLAN_ATTR_AUTH;
	}

	if ((fattrp->wa_valid & DLADM_WLAN_ATTR_BSSTYPE) != 0 &&
	    fattrp->wa_bsstype != attrp->wa_bsstype)
		return (B_TRUE);

	if ((fattrp->wa_valid & DLADM_WLAN_ATTR_BSSID) != 0 &&
	    memcmp(fattrp->wa_bssid.wb_bytes, attrp->wa_bssid.wb_bytes,
	    DLADM_WLAN_BSSID_LEN) != 0)
		return (B_TRUE);
append:
	nodep = malloc(sizeof (attr_node_t));
	if (nodep == NULL)
		return (B_TRUE);

	(void) memcpy(&nodep->an_attr, attrp, sizeof (dladm_wlan_attr_t));
	nodep->an_next = statep->cs_list;
	statep->cs_list = nodep;
	statep->cs_count++;

	return (B_TRUE);
}

dladm_status_t
dladm_wlan_connect(dladm_handle_t handle, datalink_id_t linkid,
    dladm_wlan_attr_t *attrp, int timeout, void *keys,
    uint_t key_count, uint_t flags, char *identity)
{
	char			ifname[MAXLINKNAMELEN];
	boolean_t		create_ibss;
	dladm_status_t		status;
	wl_linkstatus_t		wl_status;
	hrtime_t		start;

	if ((status = dladm_wlan_validate(handle, linkid)) != DLADM_STATUS_OK)
		return (status);

	if ((status = dladm_datalink_id2info(handle, linkid, NULL, NULL, NULL,
	    ifname, sizeof (ifname))) != DLADM_STATUS_OK)
		goto done;

	if ((status = do_get_linkstatus(handle, linkid, &wl_status,
	    sizeof (wl_status))) != DLADM_STATUS_OK)
		goto done;

	if (wl_status == WL_CONNECTED) {
		status = DLADM_STATUS_ISCONN;
		goto done;
	}

	if (wpa_instance_create(ifname))
		goto done;

	if (wpa_network_config(ifname, attrp, keys, &create_ibss, identity)) {
		status = DLADM_STATUS_BADVAL;
		goto done;
	}

	/*
	 * DLADM_WLAN_CONNECT_CREATEIBSS:
	 * If this flag is set and the bsstype attribute attr->wa_bsstype is
	 * set to DLADM_WLAN_BSSTYPE_IBSS:
	 * If the essid attribute attr->wa_essid is specified and there exists
	 * no WLAN on the discovered WLAN with this particular essid, an adhoc
	 * WLAN with essid equal to attr->wa_essid will be created.
	 * If the essid attribute is not specified, an adhoc WLAN with a random
	 * essid will be created, irrespective of which WLANs are available.
	 */

	/* we will manage this inside wpa_network config
	create_ibss = ((flags & DLADM_WLAN_CONNECT_CREATEIBSS) != 0 &&
	    attrp != NULL && (attrp->wa_valid & DLADM_WLAN_ATTR_BSSTYPE) != 0 &&
	    attrp->wa_bsstype == DLADM_WLAN_BSSTYPE_IBSS);

	if ((create_ibss && attrp != NULL &&
	    (attrp->wa_valid & DLADM_WLAN_ATTR_ESSID) == 0)) {
		status = do_connect(handle, linkid, buf, sizeof (buf), attrp,
		    create_ibss, keys, key_count, timeout);
		goto done;

	if (state.cs_count == 0) {
		if (!create_ibss) {
			status = DLADM_STATUS_NOTFOUND;
			goto done;
		}
	} */

	start = gethrtime();
	for (;;) {
		status = do_get_linkstatus(handle, linkid, &wl_status,
		    sizeof (wl_status));
		if (status != DLADM_STATUS_OK)
			goto done;

		if (wl_status == WL_CONNECTED)
			break;

		(void) poll(NULL, 0, DLADM_WLAN_CONNECT_POLLRATE);
		if ((timeout >= 0) && (gethrtime() - start) /
		    NANOSEC >= timeout) {
			status = DLADM_STATUS_TIMEDOUT;
			goto done;
		}
	}

done:
	if ((status != DLADM_STATUS_OK) && (status != DLADM_STATUS_ISCONN))
		(void) do_disconnect(handle, linkid, NULL, NULL);

	return (status);
}

dladm_status_t
dladm_wlan_disconnect(dladm_handle_t handle, datalink_id_t linkid)
{
	char		buf[WLDP_BUFSIZE];
	dladm_status_t	status;
	wl_linkstatus_t	wl_status;

	if ((status = dladm_wlan_validate(handle, linkid)) != DLADM_STATUS_OK)
		return (status);

	if ((status = do_get_linkstatus(handle, linkid, &wl_status,
	    sizeof (wl_status))) != DLADM_STATUS_OK)
		goto done;

	if (wl_status != WL_CONNECTED) {
		status = DLADM_STATUS_NOTCONN;
		goto done;
	}

	if ((status = do_disconnect(handle, linkid, buf, sizeof (buf)))
	    != DLADM_STATUS_OK)
		goto done;

	if ((status = do_get_linkstatus(handle, linkid, &wl_status,
	    sizeof (wl_status))) != DLADM_STATUS_OK)
		goto done;

	if (wl_status == WL_CONNECTED) {
		status = DLADM_STATUS_FAILED;
		goto done;
	}

	status = DLADM_STATUS_OK;
done:
	return (status);
}

dladm_status_t
dladm_wlan_get_linkattr(dladm_handle_t handle, datalink_id_t linkid,
    dladm_wlan_linkattr_t *attrp)
{
	wl_rssi_t		signal;
	wl_bss_type_t		bsstype;
	wl_authmode_t		authmode;
	/* used only in dladm show-wifi */
	wl_encryption_t		encryption;
	wl_rates_t		*ratesp = NULL;
	dladm_wlan_attr_t	*wl_attrp;
	dladm_status_t		status;
	wl_bssid_t		wl_bssid;
	wl_essid_t		wls;
	wl_phy_conf_t		wl_phy_conf;
	wl_linkstatus_t		wl_status;

	if (attrp == NULL)
		return (DLADM_STATUS_BADARG);

	if ((status = dladm_wlan_validate(handle, linkid)) != DLADM_STATUS_OK)
		goto done;

	(void) memset(attrp, 0, sizeof (*attrp));
	wl_attrp = &attrp->la_wlan_attr;

	if ((status = do_get_linkstatus(handle, linkid, &wl_status,
	    sizeof (wl_status))) != DLADM_STATUS_OK)
		goto done;

	attrp->la_valid |= DLADM_WLAN_LINKATTR_STATUS;
	if (wl_status != WL_CONNECTED)
		attrp->la_status = DLADM_WLAN_LINK_DISCONNECTED;
	else
		attrp->la_status = DLADM_WLAN_LINK_CONNECTED;

	/* essid */
	if ((status = dladm_wlan_get_essid(handle, linkid, &wls, sizeof (wls)))
	    != DLADM_STATUS_OK)
		goto done;
	wl_attrp->wa_essid.we_length = wls.wl_essid_length;
	(void) memcpy(wl_attrp->wa_essid.we_bytes, wls.wl_essid_essid,
	    wls.wl_essid_length);
	wl_attrp->wa_valid |= DLADM_WLAN_ATTR_ESSID;

	/* bssid */
	if ((status = dladm_wlan_get_bssid(handle, linkid, wl_bssid, sizeof (wl_bssid)))
	    != DLADM_STATUS_OK)
		goto done;
	(void) memcpy(wl_attrp->wa_bssid.wb_bytes, wl_bssid, DLADM_WLAN_BSSID_LEN);
	wl_attrp->wa_valid |= DLADM_WLAN_ATTR_BSSID;

	if (attrp->la_status == DLADM_WLAN_LINK_DISCONNECTED) {
		attrp->la_valid |= DLADM_WLAN_LINKATTR_WLAN;
		status = DLADM_STATUS_OK;
		goto done;
	}

	/* returned values used only in dladm show-wifi */
	if ((status = do_get_encryption(handle, linkid, &encryption,
	    sizeof (encryption))) != DLADM_STATUS_OK)
		goto done;

	wl_attrp->wa_valid |= DLADM_WLAN_ATTR_SECMODE;

	/* !!!! FIX ME!!!! */
	/*
	 *  this should be alligned with protocol/scan results,
	 * not types of keys used
	 */
	switch (encryption) {
	case WL_NOENCRYPTION:
		wl_attrp->wa_secmode = DLADM_WLAN_SECMODE_NONE;
		break;
	case WL_ENC_WEP:
		wl_attrp->wa_secmode = DLADM_WLAN_SECMODE_WEP;
		break;
	case WL_ENC_WPA:
		wl_attrp->wa_secmode = DLADM_WLAN_SECMODE_PSK;
		break;
	default:
		wl_attrp->wa_valid &= ~DLADM_WLAN_ATTR_SECMODE;
		break;
	}

	if ((status = do_get_signal(handle, linkid, &signal, sizeof (signal)))
	    != DLADM_STATUS_OK)
		goto done;

	wl_attrp->wa_valid |= DLADM_WLAN_ATTR_STRENGTH;
	wl_attrp->wa_strength = DLADM_WLAN_SIGNAL2STRENGTH(signal);

	ratesp = malloc(WLDP_BUFSIZE);
	if (ratesp == NULL) {
		status = DLADM_STATUS_NOMEM;
		goto done;
	}

	if ((status = do_get_rate(handle, linkid, ratesp, WLDP_BUFSIZE))
	    != DLADM_STATUS_OK)
		goto done;

	if (ratesp->wl_rates_num > 0) {
		uint_t	i, r = 0;

		for (i = 0; i < ratesp->wl_rates_num; i++) {
			if (ratesp->wl_rates_rates[i] > r)
				r = ratesp->wl_rates_rates[i];
		}
		wl_attrp->wa_speed = r;
		wl_attrp->wa_valid |= DLADM_WLAN_ATTR_SPEED;
	}

	if ((status = do_get_authmode(handle, linkid, &authmode,
	    sizeof (authmode))) != DLADM_STATUS_OK)
		goto done;

	wl_attrp->wa_valid |= DLADM_WLAN_ATTR_AUTH;

	switch (authmode) {
	case WL_OPENSYSTEM:
		wl_attrp->wa_auth = DLADM_WLAN_AUTH_OPEN;
		break;
	case WL_SHAREDKEY:
		wl_attrp->wa_auth = DLADM_WLAN_AUTH_SHARED;
		break;
	default:
		wl_attrp->wa_valid &= ~DLADM_WLAN_ATTR_AUTH;
		break;
	}

	if ((status = do_get_bsstype(handle, linkid, &bsstype,
	    sizeof (bsstype))) != DLADM_STATUS_OK)
		goto done;

	wl_attrp->wa_valid |= DLADM_WLAN_ATTR_BSSTYPE;

	switch (bsstype) {
	case WL_BSS_BSS:
		wl_attrp->wa_bsstype = DLADM_WLAN_BSSTYPE_BSS;
		break;
	case WL_BSS_IBSS:
		wl_attrp->wa_bsstype = DLADM_WLAN_BSSTYPE_IBSS;
		break;
	default:
		wl_attrp->wa_valid &= ~DLADM_WLAN_ATTR_BSSTYPE;
		break;
	}

	if ((status = do_get_mode(handle, linkid, &wl_phy_conf,
	    sizeof (wl_phy_conf))) != DLADM_STATUS_OK)
		goto done;

	wl_attrp->wa_mode = do_convert_mode(&wl_phy_conf);
	wl_attrp->wa_valid |= DLADM_WLAN_ATTR_MODE;
	if (wl_attrp->wa_mode != DLADM_WLAN_MODE_NONE)
		wl_attrp->wa_valid |= DLADM_WLAN_ATTR_MODE;

	attrp->la_valid |= DLADM_WLAN_LINKATTR_WLAN;
	status = DLADM_STATUS_OK;

done:
	free(ratesp);
	return (status);
}

/*
 * Check to see if the link is wireless.
 */
static dladm_status_t
dladm_wlan_validate(dladm_handle_t handle, datalink_id_t linkid)
{
	uint32_t	media;
	dladm_status_t	status;

	status = dladm_datalink_id2info(handle, linkid, NULL, NULL, &media,
	    NULL, 0);
	if (status == DLADM_STATUS_OK) {
		if (media != DL_WIFI)
			status = DLADM_STATUS_LINKINVAL;
	}
	return (status);
}

static boolean_t
find_val_by_name(const char *str, val_desc_t *vdp, uint_t cnt, uint_t *valp)
{
	int	i;

	for (i = 0; i < cnt; i++) {
		if (strcasecmp(str, vdp[i].vd_name) == 0) {
			*valp = vdp[i].vd_val;
			return (B_TRUE);
		}
	}
	return (B_FALSE);
}

static boolean_t
find_name_by_val(uint_t val, val_desc_t *vdp, uint_t cnt, char **strp)
{
	int	i;

	for (i = 0; i < cnt; i++) {
		if (val == vdp[i].vd_val) {
			*strp = vdp[i].vd_name;
			return (B_TRUE);
		}
	}
	return (B_FALSE);
}

/* data to string */

const char *
dladm_wlan_essid2str(dladm_wlan_essid_t *essid, char *buf)
{
	(void) snprintf(buf, essid->we_length+1, "%s", essid->we_bytes);
	return (buf);
}

const char *
dladm_wlan_bssid2str(dladm_wlan_bssid_t *bssid, char *buf)
{
	return (_link_ntoa(bssid->wb_bytes, buf, DLADM_WLAN_BSSID_LEN,
	    IFT_OTHER));
}

static const char *
dladm_wlan_val2str(uint_t val, val_desc_t *vdp, uint_t cnt, char *buf)
{
	char	*s;

	if (!find_name_by_val(val, vdp, cnt, &s))
		s = "";

	(void) snprintf(buf, DLADM_STRSIZE, "%s", s);
	return (buf);
}

/*
 * SECMODE SHOULD NEVER BE PRINTED, in known-wlan.conf and secobj.conf we'll use
 * secobj_class values, and for scan results wpa_s values
 */
const char *
dladm_wlan_secmode2str(dladm_wlan_secmode_t *secmode, char *buf)
{
	return (dladm_wlan_val2str((uint_t)*secmode, secmode_vals,
	    VALCNT(secmode_vals), buf));
}

const char *
dladm_wlan_strength2str(dladm_wlan_strength_t *strength, char *buf)
{
	return (dladm_wlan_val2str((uint_t)*strength, strength_vals,
	    VALCNT(strength_vals), buf));
}

const char *
dladm_wlan_mode2str(dladm_wlan_mode_t *mode, char *buf)
{
	return (dladm_wlan_val2str((uint_t)*mode, mode_vals,
	    VALCNT(mode_vals), buf));
}

const char *
dladm_wlan_speed2str(dladm_wlan_speed_t *speed, char *buf)
{
	(void) snprintf(buf, DLADM_STRSIZE, "%.*f", *speed % 2,
	    (float)(*speed) / 2);
	return (buf);
}

const char *
dladm_wlan_auth2str(dladm_wlan_auth_t *auth, char *buf)
{
	return (dladm_wlan_val2str((uint_t)*auth, auth_vals,
	    VALCNT(auth_vals), buf));
}

const char *
dladm_wlan_bsstype2str(dladm_wlan_bsstype_t *bsstype, char *buf)
{
	return (dladm_wlan_val2str((uint_t)*bsstype, bsstype_vals,
	    VALCNT(bsstype_vals), buf));
}

const char *
dladm_wlan_linkstatus2str(dladm_wlan_linkstatus_t *linkstatus, char *buf)
{
	return (dladm_wlan_val2str((uint_t)*linkstatus, linkstatus_vals,
	    VALCNT(linkstatus_vals), buf));
}

/* string to data */

dladm_status_t
dladm_wlan_str2essid(const char *str, dladm_wlan_essid_t *essid)
{
	if (str[0] == '\0' || strnlen(str, DLADM_WLAN_MAX_ESSID_LEN + 2) ==
	    DLADM_WLAN_MAX_ESSID_LEN + 2)
		return (DLADM_STATUS_BADARG);

	essid->we_length=strnlen(str, DLADM_WLAN_MAX_ESSID_LEN + 1);
	(void) memcpy(essid->we_bytes, str, essid->we_length);
	return (DLADM_STATUS_OK);
}

dladm_status_t
dladm_wlan_str2bssid(const char *str, dladm_wlan_bssid_t *bssid)
{
	int	len;
	uchar_t	*buf;

	buf = _link_aton(str, &len);
	if (buf == NULL)
		return (DLADM_STATUS_BADARG);

	if (len != DLADM_WLAN_BSSID_LEN) {
		free(buf);
		return (DLADM_STATUS_BADARG);
	}

	(void) memcpy(bssid->wb_bytes, buf, len);
	free(buf);
	return (DLADM_STATUS_OK);
}

dladm_status_t
dladm_wlan_str2secmode(const char *str, dladm_wlan_secmode_t *secmode)
{
	uint_t	val;

	if (!find_val_by_name(str, secmode_vals, VALCNT(secmode_vals), &val))
		return (DLADM_STATUS_BADARG);

	*secmode = (dladm_wlan_secmode_t)val;
	return (DLADM_STATUS_OK);
}

dladm_status_t
dladm_wlan_str2strength(const char *str, dladm_wlan_strength_t *strength)
{
	uint_t	val;

	if (!find_val_by_name(str, strength_vals, VALCNT(strength_vals), &val))
		return (DLADM_STATUS_BADARG);

	*strength = (dladm_wlan_strength_t)val;
	return (DLADM_STATUS_OK);
}

dladm_status_t
dladm_wlan_str2mode(const char *str, dladm_wlan_mode_t *mode)
{
	uint_t	val;

	if (!find_val_by_name(str, mode_vals, VALCNT(mode_vals), &val))
		return (DLADM_STATUS_BADARG);

	*mode = (dladm_wlan_mode_t)val;
	return (DLADM_STATUS_OK);
}

dladm_status_t
dladm_wlan_str2speed(const char *str, dladm_wlan_speed_t *speed)
{
	*speed = (dladm_wlan_speed_t)(atof(str) * 2);
	return (DLADM_STATUS_OK);
}

dladm_status_t
dladm_wlan_str2auth(const char *str, dladm_wlan_auth_t *auth)
{
	uint_t	val;

	if (!find_val_by_name(str, auth_vals, VALCNT(auth_vals), &val))
		return (DLADM_STATUS_BADARG);

	*auth = (dladm_wlan_auth_t)val;
	return (DLADM_STATUS_OK);
}

dladm_status_t
dladm_wlan_str2bsstype(const char *str, dladm_wlan_bsstype_t *bsstype)
{
	uint_t	val;

	if (!find_val_by_name(str, bsstype_vals, VALCNT(bsstype_vals), &val))
		return (DLADM_STATUS_BADARG);

	*bsstype = (dladm_wlan_bsstype_t)val;
	return (DLADM_STATUS_OK);
}

dladm_status_t
dladm_wlan_str2linkstatus(const char *str, dladm_wlan_linkstatus_t *linkstatus)
{
	uint_t	val;

	if (!find_val_by_name(str, linkstatus_vals,
	    VALCNT(linkstatus_vals), &val)) {
		return (DLADM_STATUS_BADARG);
	}

	*linkstatus = (dladm_wlan_linkstatus_t)val;
	return (DLADM_STATUS_OK);
}

/* legacy ioctl for WL_SCAN and WL_DISASSOCIATE */

dladm_status_t
i_dladm_wlan_legacy_ioctl(dladm_handle_t handle, datalink_id_t linkid,
    wldp_t *gbuf, uint_t id, size_t len, uint_t cmd, size_t cmdlen)
{
	char			linkname[MAXPATHLEN];
	int			fd, rc;
	struct	strioctl	stri;
	uint32_t		flags;
	dladm_status_t		status;
	uint32_t		media;
	char			link[MAXLINKNAMELEN];

	if ((status = dladm_datalink_id2info(handle, linkid, &flags, NULL,
	    &media, link, MAXLINKNAMELEN)) != DLADM_STATUS_OK) {
		return (status);
	}

	if (media != DL_WIFI)
		return (DLADM_STATUS_BADARG);

	if (!(flags & DLADM_OPT_ACTIVE))
		return (DLADM_STATUS_TEMPONLY);

	/*
	 * dlpi_open() is not used here because libdlpi depends on libdladm,
	 * and we do not want to introduce recursive dependencies.
	 */
	(void) snprintf(linkname, MAXPATHLEN, "/dev/net/%s", link);
	if ((fd = open(linkname, O_RDWR)) < 0)
		return (dladm_errno2status(errno));

	gbuf->wldp_type = NET_802_11;
	gbuf->wldp_id	= id;
	gbuf->wldp_length = len;

	stri.ic_timout	= 0;
	stri.ic_dp	= (char *)gbuf;
	stri.ic_cmd	= cmd;
	stri.ic_len	= cmdlen;

	if ((rc = ioctl(fd, I_STR, &stri)) != 0) {
		if (rc > 0) {
			/*
			 * Non-negative return value indicates the specific
			 * operation failed and the reason for the failure
			 * was stored in gbuf->wldp_result.
			 */
			status = dladm_wlan_wlresult2status(gbuf);
		} else {
			/*
			 * Negative return value indicates the ioctl failed.
			 */
			status = dladm_errno2status(errno);
		}
	}
	(void) close(fd);
	return (status);
}

static dladm_status_t
do_cmd_ioctl(dladm_handle_t handle, datalink_id_t linkid, void *buf,
    int buflen, uint_t cmd)
{
	wldp_t *gbuf;
	dladm_status_t status = DLADM_STATUS_OK;

	if ((gbuf = malloc(MAX_BUF_LEN)) == NULL)
		return (DLADM_STATUS_NOMEM);

	(void) memset(gbuf, 0, MAX_BUF_LEN);
	status = i_dladm_wlan_legacy_ioctl(handle, linkid, gbuf, cmd,
	    WLDP_BUFSIZE, WLAN_COMMAND, sizeof (wldp_t));
	(void) memcpy(buf, gbuf->wldp_buf, buflen);
	free(gbuf);
	return (status);
}

static dladm_status_t
do_scan(dladm_handle_t handle, datalink_id_t linkid, void *buf, int buflen)
{
	return (do_cmd_ioctl(handle, linkid, buf, buflen, WL_SCAN));
}

static dladm_status_t
do_disconnect(dladm_handle_t handle, datalink_id_t linkid, void *buf,
    int buflen)
{
	char ifname[MAXLINKNAMELEN];
	if ((dladm_datalink_id2info(handle, linkid, NULL, NULL, NULL, ifname,
	    sizeof (ifname))) != DLADM_STATUS_OK)
	    		return DLADM_STATUS_FAILED;
		(void) wpa_instance_delete(ifname);

	return (do_cmd_ioctl(handle, linkid, buf, buflen, WL_DISASSOCIATE));
}

/*
 * 'net80211' ioctls wrappers
 * wlan properties are not currenlty merged into the Brussels framework.
 * we should implement a common interface like this:http://pastebin.com/B05BkQJ2
 */

dladm_status_t
dladm_wlan_get_esslist(dladm_handle_t handle, datalink_id_t linkid, void *buf,
    int buflen)
{
	return (i_dladm_wlan_param(handle, linkid, buf, MAC_PROP_WL_ESS_LIST,
	    buflen, B_FALSE));
}

dladm_status_t
dladm_wlan_get_bssid(dladm_handle_t handle, datalink_id_t linkid, void *buf, int buflen)
{
	return (i_dladm_wlan_param(handle, linkid, buf, MAC_PROP_WL_BSSID,
	    buflen, B_FALSE));
}

dladm_status_t
dladm_wlan_get_essid(dladm_handle_t handle, datalink_id_t linkid, void *buf, int buflen)
{
	return (i_dladm_wlan_param(handle, linkid, buf, MAC_PROP_WL_ESSID,
	    buflen, B_FALSE));
}

static dladm_status_t
do_get_bsstype(dladm_handle_t handle, datalink_id_t linkid, void *buf,
    int buflen)
{
	return (i_dladm_wlan_param(handle, linkid, buf, MAC_PROP_WL_BSSTYPE,
	    buflen, B_FALSE));
}

static dladm_status_t
do_get_linkstatus(dladm_handle_t handle, datalink_id_t linkid, void *buf,
    int buflen)
{
	return (i_dladm_wlan_param(handle, linkid, buf, MAC_PROP_WL_LINKSTATUS,
	    buflen, B_FALSE));
}

static dladm_status_t
do_get_rate(dladm_handle_t handle, datalink_id_t linkid, void *buf, int buflen)
{
	return (i_dladm_wlan_param(handle, linkid, buf,
	    MAC_PROP_WL_DESIRED_RATES, buflen, B_FALSE));
}

static dladm_status_t
do_get_authmode(dladm_handle_t handle, datalink_id_t linkid, void *buf,
    int buflen)
{
	return (i_dladm_wlan_param(handle, linkid, buf,
	    MAC_PROP_WL_AUTH_MODE, buflen, B_FALSE));
}

static dladm_status_t
do_get_encryption(dladm_handle_t handle, datalink_id_t linkid, void *buf,
    int buflen)
{
	return (i_dladm_wlan_param(handle, linkid, buf,
	    MAC_PROP_WL_ENCRYPTION, buflen, B_FALSE));
}

static dladm_status_t
do_get_signal(dladm_handle_t handle, datalink_id_t linkid, void *buf,
    int buflen)
{
	return (i_dladm_wlan_param(handle, linkid, buf, MAC_PROP_WL_RSSI,
	    buflen, B_FALSE));
}

static dladm_status_t
do_get_mode(dladm_handle_t handle, datalink_id_t linkid, void *buf, int buflen)
{
	return (i_dladm_wlan_param(handle, linkid, buf, MAC_PROP_WL_PHY_CONFIG,
	    buflen, B_FALSE));
}

dladm_status_t
dladm_wlan_set_bsstype(dladm_handle_t handle, datalink_id_t linkid,
    dladm_wlan_bsstype_t *bsstype)
{
	wl_bss_type_t	ibsstype;

	switch (*bsstype) {
	case DLADM_WLAN_BSSTYPE_BSS:
		ibsstype = WL_BSS_BSS;
		break;
	case DLADM_WLAN_BSSTYPE_IBSS:
		ibsstype = WL_BSS_IBSS;
		break;
	default:
		ibsstype = WL_BSS_BSS;
		break;
	}
	return (i_dladm_wlan_param(handle, linkid, &ibsstype,
	    MAC_PROP_WL_BSSTYPE, sizeof (ibsstype), B_TRUE));
}

dladm_status_t
dladm_wlan_set_authmode(dladm_handle_t handle, datalink_id_t linkid,
    dladm_wlan_auth_t *auth)
{
	wl_authmode_t	auth_mode;

	switch (*auth) {
	case DLADM_WLAN_AUTH_NONE:
		return (DLADM_STATUS_OK);
	case DLADM_WLAN_AUTH_OPEN:
		auth_mode = WL_OPENSYSTEM;
		break;
	case DLADM_WLAN_AUTH_SHARED:
		auth_mode = WL_SHAREDKEY;
		break;
	default:
		return (DLADM_STATUS_NOTSUP);
	}
	return (i_dladm_wlan_param(handle, linkid, &auth_mode,
	    MAC_PROP_WL_AUTH_MODE, sizeof (auth_mode), B_TRUE));
}

/* should be renamed to a more appropriate name since it is only for WEP !!*/
dladm_status_t
dladm_wlan_set_encryption(dladm_handle_t handle, datalink_id_t linkid,
    dladm_wlan_secmode_t *secmode)
{
	wl_encryption_t	encryption;

	switch (*secmode) {
	case DLADM_WLAN_SECMODE_NONE:
		encryption = WL_NOENCRYPTION;
		break;
	case DLADM_WLAN_SECMODE_WEP:
		encryption = WL_ENC_WEP;
		break;
	case DLADM_WLAN_SECMODE_PSK:
		return (0);
	default:
		return (DLADM_STATUS_NOTSUP);
	}
	return (i_dladm_wlan_param(handle, linkid, &encryption,
	    MAC_PROP_WL_ENCRYPTION, sizeof (encryption), B_TRUE));
}

static dladm_status_t
do_set_key(dladm_handle_t handle, datalink_id_t linkid, dladm_wlan_key_t *keys,
    uint_t key_count)
{
	int			i;
	wl_wep_key_t		*wkp;
	wl_wep_key_tab_t	wepkey_tab;
	dladm_wlan_key_t	*kp;

	if (key_count == 0 || key_count > MAX_NWEPKEYS || keys == NULL)
		return (DLADM_STATUS_BADARG);

	(void) memset(wepkey_tab, 0, sizeof (wepkey_tab));
	for (i = 0; i < MAX_NWEPKEYS; i++)
		wepkey_tab[i].wl_wep_operation = WL_NUL;

	for (i = 0; i < key_count; i++) {
		kp = &keys[i];
		if (kp->wk_idx == 0 || kp->wk_idx > MAX_NWEPKEYS)
			return (DLADM_STATUS_BADARG);
		if (kp->wk_len != DLADM_WLAN_WEPKEY64_LEN &&
		    kp->wk_len != DLADM_WLAN_WEPKEY128_LEN)
			return (DLADM_STATUS_BADARG);

		wkp = &wepkey_tab[kp->wk_idx - 1];
		wkp->wl_wep_operation = WL_ADD;
		wkp->wl_wep_length = kp->wk_len;
		(void) memcpy(wkp->wl_wep_key, kp->wk_val, kp->wk_len);
	}

	return (i_dladm_wlan_param(handle, linkid, &wepkey_tab,
	    MAC_PROP_WL_KEY_TAB, sizeof (wepkey_tab), B_TRUE));
}

dladm_status_t
dladm_wlan_set_essid(dladm_handle_t handle, datalink_id_t linkid,
    const uint8_t *essid, size_t essid_len)
{
	wl_essid_t	iessid;

	(void) memset(&iessid, 0, essid_len);

	if (essid != NULL && essid_len != 0) {
		iessid.wl_essid_length = essid_len;
		(void) memcpy(iessid.wl_essid_essid, essid,
		    iessid.wl_essid_length);
	} else {
		return (DLADM_STATUS_BADARG);
	}
	return (i_dladm_wlan_param(handle, linkid, &iessid, MAC_PROP_WL_ESSID,
	    sizeof (iessid), B_TRUE));
}

dladm_status_t
do_set_bssid(dladm_handle_t handle, datalink_id_t linkid,
    dladm_wlan_bssid_t *bssid)
{
	wl_bssid_t	ibssid;

	if (bssid != NULL && bssid->wb_bytes != 0) {
		(void) memcpy(ibssid, bssid->wb_bytes, DLADM_WLAN_BSSID_LEN);
	} else {
		return (DLADM_STATUS_BADARG);
	}
	return (i_dladm_wlan_param(handle, linkid, &ibssid, MAC_PROP_WL_BSSID,
	    sizeof (ibssid), B_TRUE));
}

dladm_status_t
dladm_wlan_set_channel(dladm_handle_t handle, datalink_id_t linkid,
    dladm_wlan_channel_t *channel)
{
	wl_phy_conf_t phy_conf;

	if (*channel > MAX_CHANNEL_NUM)
		return (DLADM_STATUS_BADVAL);

	(void) memset(&phy_conf, 0xff, sizeof (phy_conf));
	phy_conf.wl_phy_dsss_conf.wl_dsss_channel = *channel;

	return (i_dladm_wlan_param(handle, linkid, &phy_conf,
	    MAC_PROP_WL_PHY_CONFIG, sizeof (phy_conf), B_TRUE));
}

dladm_status_t
do_set_createibss(dladm_handle_t handle, datalink_id_t linkid,
    boolean_t *create_ibss)
{
	wl_create_ibss_t cr = (wl_create_ibss_t)(*create_ibss);

	return (i_dladm_wlan_param(handle, linkid, &cr, MAC_PROP_WL_CREATE_IBSS,
	    sizeof (cr), B_TRUE));
}

static void
generate_essid(dladm_wlan_essid_t *essid)
{
	char ssid_temp[DLADM_WLAN_MAX_ESSID_LEN];
	srandom(gethrtime());
	snprintf(ssid_temp, DLADM_WLAN_MAX_ESSID_LEN, "Illumos-%s", random());
	(void) memcpy(essid->we_bytes,ssid_temp,16);
	essid->we_length=16;
}

dladm_status_t
do_get_capability(dladm_handle_t handle, datalink_id_t linkid, void *buf,
    int buflen)
{
	return (i_dladm_wlan_param(handle, linkid, buf, MAC_PROP_WL_CAPABILITY,
	    buflen, B_FALSE));
}

static dladm_status_t
do_get_wpamode(dladm_handle_t handle, datalink_id_t linkid, void *buf,
    int buflen)
{
	return (i_dladm_wlan_param(handle, linkid, buf, MAC_PROP_WL_WPA, buflen,
	    B_FALSE));
}

/* WPA support routines */

dladm_status_t
dladm_wlan_wpa_set_ie(dladm_handle_t handle, datalink_id_t linkid,
    const uint8_t *wpa_ie, size_t wpa_ie_len)
{
	wl_wpa_ie_t *ie;
	uint_t len;
	dladm_status_t	status;

	if (wpa_ie_len > DLADM_WLAN_MAX_WPA_IE_LEN)
		return (DLADM_STATUS_BADARG);
	len = sizeof (wl_wpa_ie_t) + wpa_ie_len;
	ie = malloc(len);
	if (ie == NULL)
		return (DLADM_STATUS_NOMEM);

	(void) memset(ie, 0, len);
	ie->wpa_ie_len = wpa_ie_len;
	(void) memcpy(ie->wpa_ie, wpa_ie, wpa_ie_len);

	status = i_dladm_wlan_param(handle, linkid, ie, MAC_PROP_WL_SETOPTIE,
	    len, B_TRUE);
	free(ie);

	return (status);
}

dladm_status_t
dladm_wlan_wpa_set_wpa(dladm_handle_t handle, datalink_id_t linkid,
    boolean_t flag)
{
	wl_wpa_t	wpa;

	wpa.wpa_flag = flag;
	return (i_dladm_wlan_param(handle, linkid, &wpa, MAC_PROP_WL_WPA,
	    sizeof (wpa), B_TRUE));
}

dladm_status_t
dladm_wlan_wpa_del_key(dladm_handle_t handle, datalink_id_t linkid,
    uint_t key_idx, const dladm_wlan_bssid_t *addr)
{
	wl_del_key_t	wk;

	wk.idk_keyix = key_idx;
	if (addr != NULL)
		(void) memcpy((char *)wk.idk_macaddr, addr->wb_bytes,
		    DLADM_WLAN_BSSID_LEN);

	return (i_dladm_wlan_param(handle, linkid, &wk, MAC_PROP_WL_DELKEY,
	    sizeof (wk), B_TRUE));
}

dladm_status_t
dladm_wlan_wpa_set_key(dladm_handle_t handle, datalink_id_t linkid,
    dladm_wlan_cipher_t cipher, const dladm_wlan_bssid_t *addr,
    boolean_t set_tx, uint64_t seq, uint_t key_idx, uint8_t *key,
    uint_t key_len)
{
	wl_key_t	wk;

	(void) memset(&wk, 0, sizeof (wl_key_t));
	switch (cipher) {
	case DLADM_WLAN_CIPHER_WEP:
		wk.ik_type = IEEE80211_CIPHER_WEP;
		break;
	case DLADM_WLAN_CIPHER_TKIP:
		wk.ik_type = IEEE80211_CIPHER_TKIP;
		break;
	case DLADM_WLAN_CIPHER_AES_OCB:
		wk.ik_type = IEEE80211_CIPHER_AES_OCB;
		break;
	case DLADM_WLAN_CIPHER_AES_CCM:
		wk.ik_type = IEEE80211_CIPHER_AES_CCM;
		break;
	case DLADM_WLAN_CIPHER_CKIP:
		wk.ik_type = IEEE80211_CIPHER_CKIP;
		break;
	case DLADM_WLAN_CIPHER_NONE:
		wk.ik_type = IEEE80211_CIPHER_NONE;
		break;
	default:
		return (DLADM_STATUS_BADARG);
	}
	wk.ik_flags = IEEE80211_KEY_RECV;
	if (set_tx) {
		wk.ik_flags |= IEEE80211_KEY_XMIT | IEEE80211_KEY_DEFAULT;
		(void) memcpy(wk.ik_macaddr, addr->wb_bytes,
		    DLADM_WLAN_BSSID_LEN);
	} else
		(void) memset(wk.ik_macaddr, 0, DLADM_WLAN_BSSID_LEN);
	wk.ik_keyix = key_idx;
	wk.ik_keylen = key_len;
	(void) memcpy(&wk.ik_keyrsc, &seq, 6);	/* only use 48-bit of seq */
	(void) memcpy(wk.ik_keydata, key, key_len);

	return (i_dladm_wlan_param(handle, linkid, &wk, MAC_PROP_WL_KEY,
	    sizeof (wk), B_TRUE));
}

dladm_status_t
dladm_wlan_wpa_set_mlme(dladm_handle_t handle, datalink_id_t linkid,
    dladm_wlan_mlme_op_t op, dladm_wlan_reason_t reason, const uint8_t *bssid)
{
	wl_mlme_t mlme;

	(void) memset(&mlme, 0, sizeof (wl_mlme_t));
	switch (op) {
	case DLADM_WLAN_MLME_ASSOC:
		mlme.im_op = IEEE80211_MLME_ASSOC;
		break;
	case DLADM_WLAN_MLME_DISASSOC:
		mlme.im_op = IEEE80211_MLME_DISASSOC;
		break;
	default:
		return (DLADM_STATUS_BADARG);
	}
	mlme.im_reason = reason;
	if (bssid != NULL)
	    memcpy(mlme.im_macaddr, bssid, DLADM_WLAN_BSSID_LEN);

	return (i_dladm_wlan_param(handle, linkid, &mlme, MAC_PROP_WL_MLME,
	    sizeof (mlme), B_TRUE));
}

 /*
  * This routine is used for opening a control interface to wpa_supplicant
  * global interface ctrl_path, /var/run/wpa_supplicant-global. This path
  * is configured in network/wpa_supplicant service manifest.
  * network/wpa_supplicant service must be running.
  */
static int
wpa_instance_create(char *ifname)
{
	struct wpa_ctrl *ctrl_global;
	const char *global_path = CTRL_IFACE_GLOBAL;

	/*
	 * When we'll have a configuration backend that supports know_wlans.conf
	 * we could add this file direclty in the third parameter
	 */
	char *interface_add[] = {"interface_add", ifname, """", "solaris",
	    CTRL_IFACE_DIR};

	if (ifname == NULL)
		return -1;

	ctrl_global = wpa_ctrl_open(global_path);
	if (ctrl_global == NULL) {
		return -1;
	}

	if (wpa_request(ctrl_global, 5, interface_add))
		return -1;

	wpa_ctrl_close(ctrl_global);

	/* wpa_s now executes wpa_driver_solaris_init */

	return 0;
}

static int
wpa_instance_delete(char *ifname)
{
	/*struct wpa_ctrl *ctrl_conn;*/
	struct wpa_ctrl *ctrl_global;

	const char *global_path = CTRL_IFACE_GLOBAL;
	/*char *cfile = NULL;*/
	char *interface_remove[] = {"interface_remove", ifname};
	/* char *cmd_disconnect[] = {"disconnect"}; */
	/* int flen, res;*/

	if (ifname == NULL)
		return -1;

/*

	 * Get the instance name of the existing control interface.
	 * Control interface will use the link name as the UNIX socket filename
	 * in CTRL_IFACE_DIR

	if (cfile == NULL) {
		flen = strlen(CTRL_IFACE_DIR) + strlen(ifname) + 2;
		cfile = malloc(flen);
		if (cfile == NULL)
			return -1;
		res = snprintf(cfile, flen, "%s/%s", CTRL_IFACE_DIR, ifname);
		if (res < 0 || res >= flen) {
			free(cfile);
			return -1;
		}
	}

	ctrl_conn = wpa_ctrl_open(cfile);

	free(cfile);

	if (ctrl_conn == NULL)
		return -1;

	if (wpa_request(ctrl_conn, 1, cmd_disconnect))
		return -1;

	wpa_ctrl_close(ctrl_conn);
*/

	/* remove interface */
	ctrl_global = wpa_ctrl_open(global_path);
	if (ctrl_global == NULL) {
		return -1;
	}

	if (wpa_request(ctrl_global, 2, interface_remove))
		return -1;

	wpa_ctrl_close(ctrl_global);

	return 0;
}

#define MAC2STR(a) (a)[0], (a)[1], (a)[2], (a)[3], (a)[4], (a)[5]
#define MACSTR "%02x:%02x:%02x:%02x:%02x:%02x"

static int wpa_network_config(char *ifname, dladm_wlan_attr_t *attrp,
	dladm_wlan_key_t *keys, boolean_t *create_ibss, char *identity)
{
	struct wpa_ctrl *ctrl_conn;

	char *cfile = NULL;
	int i, flen, res;

	char *ap_scan[] = {"ap_scan", "1"};
	char *add_network[] = {"add_network"};
	char *enable_network[] = {"enable_network", "0"};
	char *set_network[] = {"set_network", "0", NULL, NULL };
	char *network_props[] = { "ssid", "key_mgmt", "pairwise", "group",
	    "auth_alg", "proto"};
	char *network_vals[6];
	char *hexascii = NULL;
	char ap_bssid[DLADM_WLAN_BSSID_LEN*3];
	char *cmd_bssid[] = { "bssid", "0", ap_bssid };

	if (ifname == NULL)
		return -1;

	for (i=0; i<6; i++)
		network_vals[1] = NULL;

	/*
	 * Control interface will use the link name as the UNIX socket filename
	 * in CTRL_IFACE_DIR
	 */
	if (cfile == NULL) {
		flen = strlen(CTRL_IFACE_DIR) + strlen(ifname) + 2;
		cfile = malloc(flen);
		if (cfile == NULL)
			return -1;
		res = snprintf(cfile, flen, "%s/%s", CTRL_IFACE_DIR, ifname);
		if (res < 0 || res >= flen) {
			free(cfile);
			return -1;
		}
	}

	ctrl_conn = wpa_ctrl_open(cfile);
	free(cfile);

	if (ctrl_conn == NULL)
		return -1;

	/* set ap_scan=2 */
	if (wpa_request(ctrl_conn, 2, ap_scan))
		return -1;

	/* add empty wifi network block */
	(void) wpa_request(ctrl_conn, 1, add_network);

	/* leave bssid out if strict_bssid=false?*/
	snprintf(ap_bssid, 3 * DLADM_WLAN_BSSID_LEN, MACSTR,
	    MAC2STR(attrp->wa_bssid.wb_bytes));
	(void) wpa_request(ctrl_conn, 3, cmd_bssid);

	/* ssid string needs to be quoted */
	network_vals[0] = malloc(attrp->wa_essid.we_length + 3);
	snprintf(network_vals[0], attrp->wa_essid.we_length + 3, "\"%s\"",
	    attrp->wa_essid.we_bytes);
	/* we will add support for static WEP config. with wpa_s later */
	network_vals[4] = (char*)dladm_wlan_auth2str(&attrp->wa_auth, malloc(5));

	/*
	if (*create_ibss) {
		network_vals[1]="WPA-NONE";
		network_vals[2]="NONE";
		network_vals[3]="TKIP";
		network_vals[5]="WPA";
		 * mode=1
		 * frequency=2412
		 * psk="secret passphrase"
		 * ssid="test adhoc"
	} else {
		network_vals[1] = wpa_key_mgmt_txt(attrp->wa_ie.key_mgmt);
		network_vals[2] = wpa_cipher_txt(attrp->wa_ie.pairwise_cipher);
		network_vals[3] = wpa_cipher_txt(attrp->wa_ie.group_cipher);
		network_vals[5] = ((attrp->wa_ie.proto == 2) ? "RSN" : "WPA");
	}

	for (i=0; i<6; i++) {
		set_network[2] = network_props[i];
		set_network[3] = network_vals[i];
		(void) wpa_request(ctrl_conn, 4, set_network);
	} */

	set_network[2] = network_props[0];
	set_network[3] = network_vals[0];
	wpa_request(ctrl_conn, 4, set_network);
	free(network_vals[0]);
	free(network_vals[4]);

	/*
	 * If we use hexascii strings, wpa_s does not require quotation, and
	 * we would be able to pass it directly to wpa_request.
	 * Currently getsecobj returns the string value, not hexascii
	 */
	hexascii = malloc(keys->wk_len + 3);
	snprintf(hexascii, keys->wk_len + 3, "\"%s\"", keys->wk_val);

	if (keys->wk_class == DLADM_SECOBJ_CLASS_PSK) {
		set_network[2] = "psk";
		set_network[3] = hexascii;
		(void) wpa_request(ctrl_conn, 4, set_network);
	} else {
		char *cmd_eap[] = {"set_network", "0", "eap", NULL};
		set_network[2] = "identity";
		set_network[3] = identity;
		(void) wpa_request(ctrl_conn, 4, set_network);
		set_network[2] = "ca_cert";
		set_network[3] = "\"/etc/cert/ca.pem\"";
		(void) wpa_request(ctrl_conn, 4, set_network);
		 */
		if (keys->wk_class == DLADM_SECOBJ_CLASS_TLS) {
			char *tls_props[] = {"engine", "engine_id", "key_id",
			    "pin", "client_cert"};
			char *tls_vals[] = {"1", "pkcs11", "TODO", hexascii,
			    "DEFAULT_PATH"};
			cmd_eap[3] = "TLS";
			(void) wpa_request(ctrl_conn, 4, cmd_eap);
			for (i=0; i<6; i++) {
				char *cmd_tls[] = {"set_network", "0",
				    tls_props[i], tls_vals[i]};
				(void) wpa_request(ctrl_conn, 4, cmd_tls);
			}
		} else {
			set_network[2] = "password";
		    	set_network[3] = hexascii;
			(void) wpa_request(ctrl_conn, 4, set_network);
			if (keys->wk_class == DLADM_SECOBJ_CLASS_TTLS) {
				cmd_eap[3] = "TTLS";
				(void) wpa_request(ctrl_conn, 4, cmd_eap);
				/* optional
				set_network[2] = "anonymous_identity";
				set_network[3] = "DEFAULT_ANONYMOUS";
				(void) wpa_request(ctrl_conn, 4, set_network);*/
				set_network[2] = "phase2";
		    		set_network[3] = "\"auth=PAP\"";
		    		/*
		    		 * other possible values are: auth=MSCHAPV2 auth=MD5;
				 * we do not support autheap=TLS for eap-ttls
				 */
				(void) wpa_request(ctrl_conn, 4, set_network);
			} else {
				cmd_eap[3] = "PEAP";
				(void) wpa_request(ctrl_conn, 4, cmd_eap);
				set_network[2] = "phase2";
		    		set_network[3] = "\"auth=MSCHAPV2\"";
				(void) wpa_request(ctrl_conn, 4, set_network);
			}
		}
	}
	(void) wpa_request(ctrl_conn, 2, enable_network);
	/*
	 * the wpa_s driver interface starts interacting with net80211 module
	 */

	wpa_ctrl_close(ctrl_conn);

	free(hexascii);

	return 0;
}
